# Copyright (C) 2020  Oakes, Gregory <gregoryoakes@fastmail.com>
# Author: Oakes, Gregory <gregory.oakes@fastmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

__version__ = "0.1.0"


from datetime import datetime
from enum import Enum
from logging import ERROR
from subprocess import PIPE, Popen
from sys import platform
from threading import Thread

from serde import Model, fields
from serde.exceptions import SerdeError

try:
    from shutil import which
except ImportError:
    from backports.shutil_which import which

try:
    from subprocess import DEVNULL
except ImportError:
    import os

    DEVNULL = open(os.devnull, "w")


class JournalUser(Enum):
    """The journalctl perspective to use."""

    System = "system"
    User = "user"
    AllVisible = "all visible"


class JournalTime(Enum):
    """An enumeration of the possible inputs for JournalReader.since and .until."""

    PositiveRelative = "positive relative time"
    NegativeRelative = "negative relative time"
    Absolute = "absolute time"


class JournalReader(object):
    """A synchronous reader for journalctl's output."""

    def __init__(self, executable=None):
        """
        Arguments:
            executable (Optional[str]) - The path to the journalctl executable. If None,
                then the system PATH will be used to look it up.

        Raises:
            EnvironmentError - On an unsupported platform (only supported on Linux).
        """
        if not platform.startswith("linux"):
            raise EnvironmentError("Unsupported platform: {}".format(platform))
        self.executable = executable or which("journalctl")
        self.filters = []
        """A list of functions which filter records.

        The function is expected to match the following signature:

        filterer(record: JournalRecord) -> bool-like

        where False means the record should be filtered out.
        """
        self.severity_measure = lambda r: r.priority * -10 + 70
        """The function which determines severity of messages.

        The function is expected to match the following signature:

        severity_measure(record: JournalRecord) -> int

        The default is to convert the priority to logging's level by multiplying by -10
        and adding 70.
        """
        self.logger = None
        """An optional logger to which will be written records of a minimum severity."""
        self.logger_min_severity = None
        """The minimum severity to cause a record to be logged."""
        self.formatter = None
        """The format string to which the record __dict__ will be applied."""
        self.user = JournalUser.AllVisible
        """The journalctl perspective from which to pull logs."""
        self.since_type = None
        """The enumeration of possible ways in which self.since will be used."""
        self.since = None
        """The time or boot from when to pull logs.

        Can be either a datetime object or the special meaning "now".
        """
        self.until_type = None
        """The enumeration of possible ways in which self.until will be used."""
        self.until = None
        """The time or boot from when to pull logs.

        Can be either a datetime object or the special meaning "now".
        """
        self.boot = None
        """The boot from which logs should be pulled.

        Can be either an int or the special meaning "all".
        """
        self.dmesg = False
        """Whether or not to match only dmesg records.

        Warning:
        This could potentially lose records for self.since because --dmesg implies
        --boot.
        """

    def with_filter(self, new_filter):
        """Add the filter function to this monitor."""
        self.filters.append(new_filter)
        return self

    def with_severity_measure(self, severity_measure):
        """Add the severity measure function to this monitor."""
        self.severity_measure = severity_measure
        return self

    def with_logger(self, logger, severity=ERROR, formatter="{message}"):
        """Add the logger to which records of a minimum severity will be written."""
        self.logger = logger
        self.logger_min_severity = severity
        self.formatter = formatter
        return self

    def with_user(self, user):
        """Add the user for which logs should be pulled.

        Arguments:
            user (JournalUser) - The user for which logs should be pulled.
        """
        if isinstance(user, str):
            self.user = JournalUser(user)
        else:
            self.user = user
        return self

    def since(self, since, since_type=JournalTime.Absolute):
        """Only pull logs since the given time or boot."""
        if isinstance(since_type, str):
            since_type = JournalTime(since_type)
        if isinstance(since, datetime) or since == "now":
            self.since_type = since_type
            self.since = since
            return self
        raise ValueError("Unexpected value of since: {}".format(since))

    def until(self, until, until_type=JournalTime.Absolute):
        """Only pull logs until the given time or boot."""
        if isinstance(until_type, str):
            until_type = JournalTime(until_type)
        if isinstance(until, datetime) or until == "now":
            self.until_type = until_type
            self.until = until
            return self
        raise ValueError("Unexpected value of until: {}".format(until))

    def for_boot(self, boot):
        """Only pull logs for the specified boot.

        Arguments:
            boot (Union[int, str]) - Either the an integer specifying which boot for
                which logs should be pulled or the special meaning "all".
        """
        if isinstance(boot, int) or boot == "all":
            self.boot = boot
            return self
        raise ValueError("Unexpected value of boot: {}".format(boot))

    def only_dmesg(self):
        """Tell journalctl to only pull kernel messages."""
        self.dmesg = True
        return self

    @property
    def args(self):
        """The command arguments which will be used to read the journal."""
        cmd = [self.executable, "--output=json", "--no-pager"]
        if self.user != JournalUser.AllVisible:
            if self.user == JournalUser.System:
                cmd.append("--system")
            elif self.user == JournalUser.User:
                cmd.append("--user")
        if self.boot is not None:
            cmd.append("--boot={}".format(self.boot))
        if self.since is not None:
            prefix = {
                JournalTime.PositiveRelative: "+",
                JournalTime.NegativeRelative: "-",
            }.get(self.since_type, "")
            cmd.append("--since={}".format(self.since))
        if self.until is not None:
            prefix = {
                JournalTime.PositiveRelative: "+",
                JournalTime.NegativeRelative: "-",
            }.get(self.until_type, "")
            cmd.append("--until={}".format(self.until))
        if self.dmesg:
            cmd.append("--dmesg")
        return cmd

    def read_journal(self):
        """Open journalctl synchronously and yield all records matching the filters."""
        proc = Popen(self.args, stdout=PIPE, stderr=DEVNULL)
        for record, _ in self.iter_records(proc.stdout):
            yield record

    def iter_records(self, log):
        """Ingest the log and yield records.

        The log is expected to be in json format (-o json).

        Arguments:
            log (Union[file, str, bytes]) - The log to ingest. Either a file handle or a
                string.

        Yields:
            Tuple[JournalRecord, int] - The record and its severity.
        """
        # Account for python 3 bytes.
        if isinstance(log, bytes):
            log = log.decode()
        if isinstance(log, str):
            lines = log.splitlines()
        else:
            lines = log
        for line in lines:
            try:
                record = JournalRecord.from_json(line)
            except SerdeError as exp:
                if self.logger:
                    self.logger.warn("Failed to decode a journalctl record: %s", line)
                continue
            if all(filt(record) for filt in self.filters):
                severity = self.severity_measure(record)
                if self.logger and self.logger_min_severity <= severity:
                    self.logger.log(
                        msg=self.formatter.format(record.__dict__), level=severity
                    )
                yield record, severity


class JournalMonitor(Thread, JournalReader):
    """Concurrent monitor for journalctl's output."""

    def __init__(self, executable=None, **kwargs):
        super(JournalMonitor, self).__init__(**kwargs)
        JournalReader.__init__(self, executable=executable)
        self.queues = []
        """A list of queue to which records of a minimum severity will be posted."""
        self.callbacks = []
        """A list of callback functions."""

    def with_queue(self, queue, severity=ERROR):
        """Add a queue to which records of a minimum severity will be posted."""
        self.queues.append((queue, severity))
        return self

    def with_callback(self, callback, severity=ERROR):
        """Add a callback for which records of a minimum severity will be called."""
        self.callbacks.append((callback, severity))
        return self

    @property
    def args(self):
        """The command arguments which will be used to read the journal."""
        args = super(JournalMonitor, self).args
        args.append("--follow")
        return args

    def run(self):
        """Monitor the output of journalctl and report any errors."""
        proc = Popen(self.args, stdout=PIPE, stderr=DEVNULL)
        for record, severity in self.iter_records(proc.stdout):
            for callback, min_severity in self.callbacks:
                if severity >= min_severity:
                    callback(record)
            for queue, min_severity in self.queues:
                if severity >= min_severity:
                    queue.put(record)


class JournalRecord(Model):
    """A single record of the journal."""

    message = fields.Optional(fields.Str(), rename="MESSAGE")
    priority = fields.Optional(
        fields.Int(normalizers=[int]), rename="PRIORITY", default=7
    )
    syslog_facility = fields.Optional(
        fields.Int(normalizers=[int]), rename="SYSLOG_FACILITY"
    )
    syslog_id = fields.Optional(fields.Str(), rename="SYSLOG_IDENTIFIER")
